# -*- coding: utf-8 -*-

try:
    import pg_util
except ImportError:
    from . import pg_util

try:
    from odoo.modules import neutralize as odoo_neutralize
except ImportError:   # < 16.0
    odoo_neutralize = False


def neuter(db_name, track_origin_dbuuid=False, for_test=True, support=False):
    """Neutralize database `db_name` by deactivating external integrations

    The effect of the neutralization depends on the purpose passed via `for_test`. When true,
    all supported external integrations will be neutralized by setting dummy credentials and
    selecting test/demo endpoints, as appropriate. If `for_test` is false, only duplicate
    database metadata will be set via ICP.

    :param str db_name: the database to neutralize (will be done in separate cursor)
    :param bool track_origin_dbuuid: if True, the ir.config_parameter `database.uuid` is copied to
                                     a second key `origin.database.uuid`, if it does not exist yet.
    :param bool for_test: if True, explicitly neutralize all integrations
    :param bool support: set `saas.db_purpose` key to 'support' when true (used by the
                         "My Databases" dashboard)
    """
    neutralizer = Neutralizer(db_name, track_origin_dbuuid=track_origin_dbuuid,
                              for_test=for_test, support=support)
    neutralizer.neuter()


class Neutralizer():
    __doc__ = neuter.__doc__

    def __init__(self, db_name, track_origin_dbuuid=False, for_test=True, support=False):
        self.db_name = db_name
        self.track_origin_dbuuid = track_origin_dbuuid
        self.for_test = for_test
        self.support = support

    def neuter(self):
        """Run the neutralization on the database."""

        with pg_util.cursor(self.db_name) as cr:
            self._neuter_saas_pre_specifics(cr)
            if self.for_test:
                self._neuter_database(cr)
            self._neuter_saas_post_specifics(cr)

    def _neuter_saas_pre_specifics(self, cr):
        cr.execute(
            """
                DO $$
                    BEGIN
                        INSERT INTO ir_config_parameter (key, value) VALUES ('saas.db_purpose', %(purpose)s);
                    EXCEPTION
                        WHEN unique_violation THEN
                            UPDATE ir_config_parameter set value = %(purpose)s where key = 'saas.db_purpose';
                    END;
                $$;
            """,
            {'purpose': 'support' if self.support else 'test' if self.for_test else ''}
        )
        # regenerate new push notification uuid
        cr.execute("DELETE FROM ir_config_parameter WHERE key='ocn.uuid'")
        # delete domains on websites
        cr.execute("""
            DO $$
                BEGIN
                    UPDATE website SET domain = null;
                EXCEPTION
                    WHEN undefined_table OR undefined_column THEN
                END;
            $$;
        """)
        if self.track_origin_dbuuid:
            cr.execute("""
                INSERT INTO ir_config_parameter(key, value)
                        SELECT 'origin.database.uuid', value
                        FROM ir_config_parameter
                        WHERE key='database.uuid'
                        AND NOT EXISTS(SELECT 1
                                            FROM ir_config_parameter
                                        WHERE key='origin.database.uuid')
            """)

    def _neuter_saas_post_specifics(self, cr):
        if self.for_test:
            # standard neuter disables all crons but autovacuum, we must ensure
            # ping and update crons also stays enabled
            cr.execute("""
                UPDATE ir_cron SET active=true
                WHERE id IN (
                    SELECT res_id FROM ir_model_data
                    WHERE model='ir.cron'
                    AND (
                            (module, name) IN (
                                ('mail', 'ir_cron_module_update_notification'),
                                ('base', 'autovacuum_job')
                            )
                            OR module = '__upgrade__'
                    )
                )
            """)

            cr.execute("""
                DO $$
                    BEGIN
                        UPDATE fetchmail_server SET active = false;
                    EXCEPTION
                        WHEN undefined_table THEN
                    END;
                $$;
            """)  # disable l10n_cl dte mailboxes as well as mailboxes in general
            cr.execute("""
                DO $$
                    BEGIN
                        UPDATE account_online_link SET provider_data = '', client_id = 'duplicate';
                    EXCEPTION
                        WHEN undefined_table OR undefined_column THEN
                    END;
                $$;
            """)  # disable OdooFin synchronization

    if odoo_neutralize:  # >= 16.0
        def _neuter_database(self, cr):
            odoo_neutralize.neutralize_database(cr)

    else:  # <= saas-15.4
        def _neuter_database(self, cr):
            # WARNING: legacy neutralization, this should only be used for versions <= saas~15.4 !!!
            cr.execute("""
                UPDATE ir_cron SET active=false
                WHERE id NOT IN (
                    SELECT res_id FROM ir_model_data
                    WHERE model='ir.cron' AND (module, name) IN (
                                ('mail', 'ir_cron_module_update_notification'),
                                ('base', 'autovacuum_job')
                        )
                    )
            """)
            cr.execute("UPDATE ir_mail_server SET active = false, smtp_port = 1025, smtp_host='localhost'")
            cr.execute("""
                DO $$
                    BEGIN
                        INSERT INTO ir_mail_server(name,smtp_port,smtp_host,smtp_encryption,active,smtp_authentication)
                                    VALUES ('neutralization - disable emails',1025,'invalid','none',true,'login');
                    EXCEPTION
                        WHEN undefined_column THEN
                            -- before v15, no `smtp_authentication` column
                            INSERT INTO ir_mail_server(name,smtp_port,smtp_host,smtp_encryption,active)
                                        VALUES ('neutralization - disable emails',1025,'invalid','none',true);
                    END;
                $$;
            """)
            cr.execute("""
                DO $$
                    BEGIN
                        UPDATE mail_template SET mail_server_id=null;
                    EXCEPTION
                        WHEN undefined_table OR undefined_column THEN
                    END;
                $$;
            """)
            cr.execute("""
                DO $$
                    BEGIN
                        UPDATE res_company SET l10n_ar_afip_ws_environment = 'testing';
                        DELETE FROM ir_attachment
                        WHERE res_model = 'res.company' and res_field in ('l10n_ar_afip_ws_crt', 'l10n_ar_afip_ws_key');
                    EXCEPTION
                        WHEN undefined_column THEN
                    END;
                $$;
            """)  # disable l10n_ar afip integration
            cr.execute("""
                DO $$
                    BEGIN
                        INSERT INTO ir_config_parameter (key, value) VALUES ('account_edi_proxy_client.demo', true);
                    EXCEPTION
                        WHEN unique_violation THEN
                            UPDATE ir_config_parameter set value = true where key = 'account_edi_proxy_client.demo';
                    END;
                $$;
            """)
            cr.execute("""
                DO $$
                    BEGIN
                        UPDATE account_edi_proxy_client_user
                        SET edi_mode = CASE
                                            WHEN proxy_type = 'l10n_it_edi' THEN 'demo'
                                            ELSE 'test'
                                        END;
                    EXCEPTION
                        WHEN undefined_table OR undefined_column THEN
                    END;
                $$;
            """)  # disable edi connection in general, and l10n_it_edi_sdicoop in particular
            cr.execute("""
                DELETE FROM ir_config_parameter
                    WHERE key='ocn.ocn_push_notification' OR
                            key='odoo_ocn.project_id';
            """)  # disable mobile push notifications
            cr.execute(r"""
                DO $$
                    BEGIN
                        UPDATE iap_account SET account_token = REGEXP_REPLACE(account_token, '(\+.*)?$', '+disabled');
                    EXCEPTION
                        WHEN undefined_table THEN
                    END;
                $$;
            """)  # disable IAP accounts

            # disbale website_event_jitsi
            cr.execute("""
                DELETE FROM ir_config_parameter
                WHERE key = 'website_jitsi.jitsi_server_domain';
            """)

            # disable l10n_in_edi_ewaybill
            cr.execute("DELETE FROM ir_config_parameter WHERE key='l10n_in_edi_ewaybill.endpoint';")

            # Web Push notification (mail_enterprise)
            cr.execute("""
                -- delete VAPID keys
                DELETE FROM ir_config_parameter
                WHERE key IN (
                       'mail_enterprise.web_push_vapid_private_key',
                       'mail_enterprise.web_push_vapid_public_key'
                       );
            """)
            cr.execute("""
                DO $$
                    BEGIN
                        -- delete delayed messages (CRON)
                        TRUNCATE mail_notification_web_push;
                        -- delete Devices for each partners
                        TRUNCATE mail_partner_device CASCADE;
                    EXCEPTION
                        WHEN undefined_table OR undefined_column THEN
                    END;
                $$;
            """)

            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  REMINDER !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # This "manual" neutralisation only needs to be done for versions prior to 16.0,
            # the standard neutralisation framework is used from 16.0 onwards.
            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  REMINDER !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

            # disable all standard payments providers credentials
            payment_acquirer_credential_columns = [
                "adyen_merchant_account",
                "adyen_skin_code",
                "adyen_skin_hmac_key",
                "adyen_api_key",
                "adyen_client_key",
                "alipay_merchant_partner_id",
                "alipay_md5_signature_key",
                "alipay_seller_email",
                "aps_merchant_identifier",
                "aps_access_code",
                "aps_sha_request",
                "aps_sha_response",
                "asiapay_merchant_id",
                "asiapay_secure_hash_secret",
                "authorize_login",
                "authorize_transaction_key",
                "authorize_signature_key",
                "authorize_client_key",
                "brq_websitekey",
                "brq_secretkey",
                "flutterwave_public_key",
                "flutterwave_secret_key",
                "flutterwave_webhook_secret",
                "mercado_pago_access_token",
                "ogone_pspid",
                "ogone_userid",
                "ogone_password",
                "ogone_shakey_in",
                "ogone_shakey_out",
                "paypal_email_account",
                "paypal_seller_account",
                "paypal_pdt_token",
                "payulatam_merchant_id",
                "payulatam_account_id",
                "payulatam_api_key",
                "payumoney_merchant_key",
                "payumoney_merchant_salt",
                "razorpay_key_id",
                "razorpay_key_secret",
                "razorpay_webhook_secret",
                "sips_merchant_id",
                "sips_secret",
                "stripe_secret_key",
                "stripe_publishable_key",
            ]

            # UPDATES = {<table>: [<updates_group>, ...]}
            # where <updates_group>  is a tuple (<column>, <value>, <condition>:optional)
            # or a list of such tuples if multiple columns need to be updated at the same time
            # note: only the first condition in an <updates_group> is taken into account
            UPDATES = {
                "account_fiscal_position": [  # Avatax (also prevent call to sandbox server)
                    ("is_avatax", False),
                ],
                "account_online_link": [  # OdooFin synchronization
                    ("provider_data", ''),
                    ("client_id", 'duplicate'),
                ],
                "account_online_provider": [  # Plaid
                    ("provider_account_identifier", None),
                ],
                "amazon_account": [
                    [
                        ("seller_key", 'dummy'),  # sale_amazon < 15.4
                        ("auth_token", 'dummy'),
                    ],
                    [
                        ("seller_key", 'dummy'),  # sale_amazon_spapi < 15.4 and sale_amazon >= 15.4
                        ("refresh_token", 'dummy'),
                    ],
                ],
                "delivery_carrier": [
                    ("prod_environment", False),
                    [
                        ("sendcloud_public_key", 'dummy'),
                        ("sendcloud_secret_key", 'dummy'),
                    ],
                ],
                "fetchmail_server": [
                    ("active", False),
                ],
                "google_calendar_credentials": [  # google_calendar >= 15
                    [
                        ("calendar_rtoken", None),
                        ("calendar_token", None),
                        ("synchronization_stopped", True),
                    ],
                ],
                "ir_config_parameter": [
                    ("value", 'demo', "key = 'voip.mode'"),  # VOIP
                    ("value", 'dummy', "key in ('%s')" % ','.join(key for key in [
                        'social.facebook_app_id',
                        'social.facebook_client_secret',
                        'social.instagram_app_id',
                        'social.instagram_client_secret',
                        'social.linkedin_app_id',
                        'social.linkedin_client_secret',
                        'social.twitter_consumer_key',
                        'social.twitter_consumer_secret_key',
                    ])),  # social media
                ],
                "payment_acquirer": [
                    ("state", 'test', "state != 'disabled'"),
                    ("environment", 'test'),  # < 13.0
                    *[(col, 'dummy') for col in payment_acquirer_credential_columns],
                    [
                        ("stripe_account_id", None),
                        ("stripe_account_details_submitted", None),
                    ],  # Defined in saas_payment_stripe
                ],
                "pos_payment_method": [  # Adyen POS integration
                    ("adyen_test_mode", True),
                ],
                "res_company": [
                    [
                        ("yodlee_access_token", None),
                        ("yodlee_user_password", None),
                        ("yodlee_user_access_token", None),
                    ],
                    ("l10n_mx_edi_pac_test_env", True),
                    ("l10n_co_edi_test_mode", True),
                    ("l10n_pe_edi_test_env", True),
                    [
                        ("l10n_pe_edi_stock_client_id", ""),
                        ("l10n_pe_edi_stock_client_secret", ""),
                        ("l10n_pe_edi_stock_client_username", ""),
                        ("l10n_pe_edi_stock_client_password", ""),
                        ("l10n_pe_edi_stock_token", ""),
                    ],
                    ("l10n_es_edi_test_env", True),
                    ("l10n_cl_dte_service_provider", 'SIITEST'),
                    ("l10n_in_edi_production_env", False),
                    [
                        ("l10n_eg_production_env", False),
                        ("l10n_eg_client_secret", 'dummy'),
                    ],
                    [
                        ("onss_expeditor_number", 'dummy'),
                        ("onss_pem_passphrase", 'dummy'),
                    ],
                    [
                        ("taxcloud_api_id", None),
                        ("taxcloud_api_key", None),
                    ],
                    [
                        ("l10n_de_fiskaly_api_key", None),
                        ("l10n_de_fiskaly_api_secret", None),
                        ("l10n_de_fiskaly_organization_id", None),
                    ],
                    ("avalara_environment", 'sandbox'),
                    ("l10n_ke_cu_proxy_address", 'dummy_url'),
                    ("l10n_in_gstr_gst_production_env", False),
                    [
                        ("l10n_in_edi_ewaybill_username", None),
                        ("l10n_in_edi_ewaybill_password", None),
                        ("l10n_in_edi_ewaybill_auth_validity", None),
                    ],
                    ("l10n_ec_production_env", False),
                    ("l10n_sa_api_mode", 'sandbox'),
                    ("l10n_br_avalara_environment", 'sandbox'),
                ],
                "res_config_settings": [
                    ("mode", 'demo'),  # VOIP
                ],
                "res_users": [
                    [
                        ("microsoft_calendar_token", None),  # microsoft_account
                        ("microsoft_calendar_rtoken", None),
                    ],
                    ("microsoft_synchronization_stopped", True),  # microsoft_calendar >= 15
                    [
                        ("google_calendar_token", None),  # google_calendar < 15
                        ("google_calendar_rtoken", None),
                    ],
                ],
                "social_account": [
                    (col, None)
                    for col in [
                        'facebook_account_id',
                        'facebook_access_token',
                        'instagram_account_id',
                        'instagram_facebook_account_id',
                        'instagram_access_token',
                        'linkedin_account_urn',
                        'linkedin_access_token',
                        'twitter_user_id',
                        'twitter_oauth_token',
                        'twitter_oauth_token_secret',
                        'youtube_channel_id',
                        'youtube_access_token',
                        'youtube_refresh_token',
                        'youtube_token_expiration_date',
                        'youtube_upload_playlist_id'
                    ]
                ],  # Posts are made via the access token stored on social_account
                # without checking the system parameters
                "website": [
                    [
                        ("firebase_enable_push_notifications", False),  # firebase
                        ("firebase_use_own_account", False),
                        ("firebase_project_id", None),
                        ("firebase_web_api_key", None),
                        ("firebase_push_certificate_key", None),
                        ("firebase_sender_id", None),
                    ],
                    ("google_places_api_key", 'dummy'),  # website_sale_autocomplete
                ],
            }

            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  REMINDER !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # This "manual" neutralisation only needs to be done for versions prior to 16.0,
            # the standard neutralisation framework is used from 16.0 onwards.
            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  REMINDER !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

            for table, updates in UPDATES.items():
                for update in updates:
                    if isinstance(update[0], str):
                        update = [update]
                    cols, vals, where_clause = [], [], ""
                    for update_line in update:
                        assert len(update_line) in (2, 3), "neutralize wrong format - (key, value[, condition])"
                        col, val = update_line[:2]
                        cols.append(col)
                        vals.append(val)
                        if not where_clause:
                            where_clause = "WHERE %s" % update_line[2] if len(update_line) > 2 else ""
                    cols_str = ', '.join('%s = %%s' % col for col in cols)
                    query = "UPDATE %s SET %s %s;" % (table, cols_str, where_clause)
                    wrapped_query = """
                        DO $$
                            BEGIN
                                %s
                            EXCEPTION
                                WHEN undefined_table OR undefined_column THEN
                            END;
                        $$;
                    """ % query
                    cr.execute(wrapped_query, vals)

            # Mark the db as neutralized. This is used by code like `ir.cron.toggle` which behaves differently if
            # it's running in a neutralized duplicate. From saas-15.2 onward this parameter is set by the new
            # neutralize framework. We're setting it here for compatibility with Odoo 13.0 to saas-15.1, and for
            # later versions until SaaS starts using the neutralize framework itself.
            cr.execute("""
                DO $$
                    BEGIN
                        INSERT INTO ir_config_parameter (key, value) VALUES ('database.is_neutralized', 'True');
                    EXCEPTION
                        WHEN unique_violation THEN
                        UPDATE ir_config_parameter set value = 'True' where key = 'database.is_neutralized';
                    END;
                $$;
            """)

            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  REMINDER !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # This "manual" neutralisation only needs to be done for versions prior to 16.0,
            # the standard neutralisation framework is used from 16.0 onwards.
            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  REMINDER !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
